import 'dart:convert';

import 'package:appwrite/appwrite.dart';
import 'package:appwrite/models.dart';
import 'package:car_app/exception/invalid_server_response.dart';
import 'package:car_app/exception/server.dart';
import 'package:flutter/foundation.dart';

class AppwriteService {
  static const String _productionEndpoint = 'https://appwrite.420joos.dev/v1';
  static const String _developmentEndpoint = 'https://appwrite.testenv.420joos.dev/v1';
  static const String _projectId = '65fcb48c714ed68ec324';
  static const String _apiFunctionId = "65fde2f5abbbe63b7804";

  static final AppwriteService _appwriteService = AppwriteService._privateConstructor();
  
  final Client _client = Client();
  late final Account _account;
  late final Functions _functions;

  factory AppwriteService() => _appwriteService;

  Client get client => _client;
  Account get account => _account;
  Functions get functions => _functions;
  String get apiFunctionId => _apiFunctionId;

  AppwriteService._privateConstructor() {
    var endpoint = _productionEndpoint;
    if (!kReleaseMode) {
      endpoint = _developmentEndpoint;
    }

    client
      .setEndpoint(endpoint)
      .setProject(_projectId);

    _account = Account(client);
    _functions = Functions(client);
  }

  bool isFunctionSuccessful(Execution execution) {
    return execution.status == "completed";
  }

  Future<Execution> executeFunction(Future<Execution> function, String action) async {
    Execution execution;
    try {
      execution = await function;
    } on AppwriteException catch (_) {
      throw ServerException(action: "$action, error while executing function");
    }

    if(!_appwriteService.isFunctionSuccessful(execution)) {
      throw ServerException(action: "$action, function was not successful");
    }

    return execution;
  }

  bool validateExecutionBody(Execution execution) {
    return execution.responseBody.isNotEmpty;
  }

  void throwIfExecutionBodyIsEmpty(Execution execution, String action) {
    if(!_appwriteService.validateExecutionBody(execution)) {
      throw InvalidServerResponseException(action: action);
    }
  }

  T fromExecution<T>(Execution execution, T Function(Map<String, dynamic>) fromMap, String action) {
    try {
      var decodedEvent = jsonDecode(execution.responseBody) as Map<String, dynamic>;
      return fromMap(decodedEvent);
    } on TypeError catch (_) {
      throw InvalidServerResponseException(action: action);
    }
  }

  Stream<T> streamFromExecution<T>(Execution execution, T Function(Map<String, dynamic>) fromMap, String action) async* {
    try {
      var decodedEvents = jsonDecode(execution.responseBody) as List<dynamic>;
      for (var event in decodedEvents) {
        yield fromMap(event);
      }
    } on TypeError catch (_) {
      throw InvalidServerResponseException(action: action);
    }
  }
}
