import 'dart:async';
import 'dart:ui';
import 'dart:io' show Platform;

import 'package:car_app/converter/location_converters.dart/point.dart';
import 'package:car_app/converter/position_point_converter.dart';
import 'package:car_app/exception/permission.dart';
import 'package:car_app/model/point.dart';
import 'package:flutter_background_service/flutter_background_service.dart';
import 'package:flutter_background_service_android/flutter_background_service_android.dart';
import 'package:geolocator/geolocator.dart';
import 'package:permission_handler/permission_handler.dart';

import 'package:car_app/converter/converter.dart';

enum Mode {
  foreground,
  background,
}

class LocationService {
  static final LocationService _locationService = LocationService._privateConstructor();
  final FlutterBackgroundService _backgroundService = FlutterBackgroundService();
  final StreamController<Point> _locationStream = StreamController.broadcast();

  late StreamSubscription _backgroundServiceSubscription;

  factory LocationService() => _locationService;

  Stream<Point> get locationUpdate => _locationStream.stream;

  LocationService._privateConstructor();

  Future<void> start() async {
    var permissionStatus = await Permission.locationWhenInUse.request();
    if (permissionStatus != PermissionStatus.granted) {
      throw PermissionException(permissions: ["location"]);
    }

    _backgroundService.configure(
      androidConfiguration: AndroidConfiguration(
        onStart: _onServiceStart,
        isForegroundMode: true,
      ),
      iosConfiguration: IosConfiguration(),
    );
    _backgroundServiceSubscription = _backgroundService.on('update').listen((event) {
      if (event != null && event.containsKey('point')) {
        var point = PointMapConverter.fromMap(event['point']);
        _locationStream.add(point);
      }
    });
    _backgroundService.startService();
  }

  static void _onServiceStart(ServiceInstance service) async {
    final Converter<Position, Point> converter = PositionToPointConverter();
    final StreamController<Position> positionUpdate = StreamController<Position>.broadcast();
    late Stream<Position> positionStream;
    final List<Position> backgroundStore = [];
    StreamSubscription<Position>? foregroundSubscription;
    StreamSubscription<Position>? backgroundSubscription;

    void sendUpdate(Position position) {
      var point = converter.convert(position);
      service.invoke('update',{
        'point': point.toMap(),
      });
    }

    LocationSettings locationSettings;
    if (Platform.isAndroid) {
      locationSettings = AndroidSettings(
        accuracy: LocationAccuracy.best,
        distanceFilter: 0,
        foregroundNotificationConfig: const ForegroundNotificationConfig(
          notificationTitle: "Location updating in background",
          notificationText: "This app will continue to receive your location even when you aren't using it",
          enableWakeLock: true,
      ));
    }
    else {
      locationSettings = const LocationSettings(
        accuracy: LocationAccuracy.best,
        distanceFilter: 0,
      );
    }

    positionStream = Geolocator.getPositionStream(
      locationSettings: locationSettings,
    );

    DartPluginRegistrant.ensureInitialized();

    var positionStreamSubscription = positionStream.listen(positionUpdate.add);
    foregroundSubscription = positionUpdate.stream.listen(sendUpdate); // at initialization 'setAsForeground' is not called

    if (service is AndroidServiceInstance) {
      service.on('setAsForeground').listen((event) async {
        await service.setAsForegroundService();
        await backgroundSubscription?.cancel();
        backgroundStore.forEach(sendUpdate);
        backgroundStore.clear();
        
        foregroundSubscription = positionUpdate.stream.listen(sendUpdate);
      });

      service.on('setAsBackground').listen((event) async {
        service.setAsBackgroundService();
        await foregroundSubscription?.cancel();
        backgroundSubscription = positionUpdate.stream.listen((position) {
          backgroundStore.add(position);
        });
      });
    }

    service.on('stop').listen((event) async {
      await service.stopSelf();
      await foregroundSubscription?.cancel();
      await backgroundSubscription?.cancel();
      await positionUpdate.close();
      positionStreamSubscription.cancel();
    });

    Timer.periodic(const Duration(seconds: 1), (timer) async {
    if (service is AndroidServiceInstance) {
        if (await service.isForegroundService()) {
          service.setForegroundNotificationInfo(
            title: "SharedCar location service",
            content: "Updated at ${DateTime.now()}",
          );
        }
      }
    });
  }

  Future<void> stop() async {
    _backgroundService.invoke('stop');
    _backgroundServiceSubscription.cancel();
  }

  void setMode(Mode mode) {
    if (mode == Mode.foreground) {
      _backgroundService.invoke('setAsForeground');
    }
    else if (mode == Mode.background) {
      _backgroundService.invoke('setAsBackground');
    }
  }

  void dispose() {
    _backgroundService.invoke('stop');
    _locationStream.close();
  }
}
