import 'package:car_app/exception/server.dart';
import 'package:car_app/model/calendar_event.dart';
import 'package:car_app/model/user.dart';
import 'package:car_app/service/interface/authentication.dart';
import 'package:car_app/service/interface/calendar_event.dart';
import 'package:car_app/service/interface/logging.dart';
import 'package:flutter/foundation.dart';

class CalendarViewModel extends ChangeNotifier {
  final CalendarEventService calendarEventService;
  final AuthenticationService authenticationService;
  final LoggingService loggingService;
  final List<CalendarEvent> events = [];

  User? currentUser;

  CalendarViewModel({required this.calendarEventService, required this.authenticationService, required this.loggingService}) {
    authenticationService.isLoggedIn().then((isLoggedIn) async {
      if(isLoggedIn) {
        currentUser = await authenticationService.user;
      }
    })
    .catchError(
      test: (e) => e is ServerException, 
      (e, s) => loggingService.error(e.toString(), exception: e, trace: s),
    );
  }

  void addNewlyAddedCalendarEvent(CalendarEvent event) {
    _addEvent(event);
  }

  void _addEvent(CalendarEvent event) {
    events.add(event);
    events.sort((a, b) => a.start.compareTo(b.start));
    notifyListeners();
  }

  void loadEvents() {
    events.clear();

    authenticationService.isLoggedIn().then((isLoggedIn) {
      if(isLoggedIn) {
        calendarEventService.getAll().forEach((event) {
          _addEvent(event);
        });
      }
    })
    .catchError(test: (e) => e is ServerException,
      (e, s) => loggingService.error(e.toString(), exception: e, trace: s),
    );
  }

  String avatarTextFromName(String name) {
    var nameParts = name.split(" ");
    if (nameParts.isEmpty) {
      return "";
    }

    var firstName = nameParts.first;
    while (firstName.isEmpty && nameParts.isNotEmpty) {
      nameParts.removeAt(0);
      firstName = nameParts.first;
    }

    var initials = "";
    if (firstName.isNotEmpty) {
      initials = firstName[0].toUpperCase();
    }

    if (nameParts.length <= 1) {
      return initials;
    }

    var lastName = nameParts.last;
    while (lastName.isEmpty && nameParts.isNotEmpty) {
      nameParts.removeLast();
      lastName = nameParts.last;
    }

    if (lastName.isNotEmpty) {
      initials += lastName[0].toUpperCase();
    }

    return initials;
  }

  bool isEventOnDay(DateTime day, CalendarEvent event) {
    var dayStart = DateTime(day.year, day.month, day.day);
    var dayEnd = DateTime(day.year, day.month, day.day + 1);
    var eventStart = event.start;
    var eventEnd = event.end;
    var startsInPast = eventStart.isBefore(dayStart);
    var endsInFuture = eventEnd.isAfter(dayEnd);
    var startsOnThisDay = (eventStart.isAfter(dayStart) || eventStart.isAtSameMomentAs(dayStart)) && eventStart.isBefore(dayEnd);
    var endsOnThisDay = eventEnd.isAfter(dayStart) && (eventEnd.isBefore(dayEnd) || eventEnd.isAtSameMomentAs(dayEnd));
    return (startsInPast && endsInFuture)
      || (startsOnThisDay && endsOnThisDay)
      || (startsOnThisDay && endsInFuture)
      || (startsInPast && endsOnThisDay);
  }

  List<CalendarEvent> eventsOfDay(DateTime day) {
    return events.where((element) => isEventOnDay(day, element)).toList();
  }

  void replaceEvent(CalendarEvent oldEvent, CalendarEvent newEvent) {
    events.remove(oldEvent);
    _addEvent(newEvent);
  }

    void deleteEvent(CalendarEvent event, void Function(CalendarEvent) undoToast) {
    if(isEventFromCurrentUser(event)) {
      calendarEventService.delete(event).then((event) {
        _removeEvent(event);
        undoToast(event);
      })
      .catchError(
        test: (e) => e is ServerException,
        (e, s) => loggingService.error(e.toString(), exception: e, trace: s),
      );
    }
  }

  void _removeEvent(CalendarEvent event) {
    events.removeWhere((e) => e.id == event.id); //events.remove(event) does not remove the event
    notifyListeners();
  }

  void addEvent(CalendarEvent event) {
    _addEvent(event); // to have an immediate feedback, event is replaced with the event from the server later
    calendarEventService.add(event).then((eventFromServer) {
      _removeEvent(event);
      _addEvent(eventFromServer);
    })
    .catchError(
      test: (e) => e is ServerException,
      (e, s) {
        loggingService.error(e.toString(), exception: e, trace: s);
        _removeEvent(event);
      },
    );
  }

  bool isEventFromCurrentUser(CalendarEvent event) {
    return currentUser != null && event.user.id == currentUser!.id;
  }

  DateTime getSelectedDateWithCurrentTime(DateTime? selectedDay) {
    var currentDateTime = DateTime.now();

    if(selectedDay == null) {
      return currentDateTime;
    }

    return DateTime(
      selectedDay.year,
      selectedDay.month,
      selectedDay.day,
      currentDateTime.hour,
      currentDateTime.minute,
      currentDateTime.second,
      currentDateTime.millisecond,
      currentDateTime.microsecond,
    );
  }
}
